@@ -1,4 +1,4 @@
-use crate::{Job, UnitOfWork, Worker};
+use crate::{Job, UnitOfWork, WorkerRef};
 use async_trait::async_trait;
 use bb8_redis::{bb8::Pool, RedisConnectionManager};
 use slog::error;
@@ -12,8 +12,8 @@ pub trait ServerMiddleware {
     async fn call(
         &self,
         iter: ChainIter,
-        job: Job,
-        worker: Box<dyn Worker>,
+        job: &Job,
+        worker: Arc<WorkerRef>,
         redis: Pool<RedisConnectionManager>,
     ) -> ServerResult;
 }
@@ -29,8 +29,8 @@ pub struct ChainIter {
 impl ChainIter {
     pub async fn next(
         &self,
-        job: Job,
-        worker: Box<dyn Worker>,
+        job: &Job,
+        worker: Arc<WorkerRef>,
         redis: Pool<RedisConnectionManager>,
     ) -> ServerResult {
         let stack = self.stack.read().await;
@@ -42,8 +42,8 @@ impl ChainIter {
                         stack: self.stack.clone(),
                         index: self.index + 1,
                     },
-                    job.clone(),
-                    worker.clone(),
+                    job,
+                    worker,
                     redis,
                 )
                 .await?;
@@ -60,6 +60,14 @@ pub(crate) struct Chain {
 }
 
 impl Chain {
+    // Testing helper to get an empty chain.
+    #[allow(dead_code)]
+    pub(crate) fn empty() -> Self {
+        Self {
+            stack: Arc::new(RwLock::new(vec![])),
+        }
+    }
+
     pub(crate) fn new(logger: slog::Logger) -> Self {
         Self {
             stack: Arc::new(RwLock::new(vec![
@@ -72,7 +80,8 @@ impl Chain {
     pub(crate) async fn using(&mut self, middleware: Box<dyn ServerMiddleware + Send + Sync>) {
         let mut stack = self.stack.write().await;
         // HACK: Insert after retry middleware but before the handler middleware.
-        let index = stack.len() - 1;
+        let index = if stack.is_empty() { 0 } else { stack.len() - 1 };
+
         stack.insert(index, middleware);
     }
 
@@ -85,8 +94,8 @@ impl Chain {
 
     pub(crate) async fn call(
         &mut self,
-        job: Job,
-        worker: Box<dyn Worker>,
+        job: &Job,
+        worker: Arc<WorkerRef>,
         redis: Pool<RedisConnectionManager>,
     ) -> ServerResult {
         // The middleware must call bottom of the stack to the top.
@@ -104,11 +113,11 @@ impl ServerMiddleware for HandlerMiddleware {
     async fn call(
         &self,
         _chain: ChainIter,
-        job: Job,
-        worker: Box<dyn Worker>,
+        job: &Job,
+        worker: Arc<WorkerRef>,
         _redis: Pool<RedisConnectionManager>,
     ) -> ServerResult {
-        worker.perform(job.args).await
+        worker.call(job.args.clone()).await
     }
 }
 
@@ -127,19 +136,21 @@ impl ServerMiddleware for RetryMiddleware {
     async fn call(
         &self,
         chain: ChainIter,
-        mut job: Job,
-        worker: Box<dyn Worker>,
+        job: &Job,
+        worker: Arc<WorkerRef>,
         mut redis: Pool<RedisConnectionManager>,
     ) -> ServerResult {
         let max_retries = worker.max_retries();
 
         let err = {
-            match chain.next(job.clone(), worker, redis.clone()).await {
+            match chain.next(job, worker, redis.clone()).await {
                 Ok(()) => return Ok(()),
                 Err(err) => format!("{err:?}"),
             }
         };
 
+        let mut job = job.clone();
+
         // Update error fields on the job.
         job.error_message = Some(err);
         if job.retry_count.is_some() {
@@ -161,11 +172,71 @@ impl ServerMiddleware for RetryMiddleware {
                 "err" => &job.error_message,
             );
 
-            UnitOfWork::from_job(job.clone())
-                .reenqueue(&mut redis)
-                .await?;
+            UnitOfWork::from_job(job).reenqueue(&mut redis).await?;
         }
 
         Ok(())
     }
 }
+
+#[cfg(test)]
+mod test {
+    use super::*;
+    use crate::Worker;
+    use tokio::sync::Mutex;
+
+    async fn redis() -> Pool<RedisConnectionManager> {
+        let manager = RedisConnectionManager::new("redis://127.0.0.1/").unwrap();
+        Pool::builder().build(manager).await.unwrap()
+    }
+
+    fn job() -> Job {
+        Job {
+            class: "TestWorker".into(),
+            queue: "default".into(),
+            args: vec![1337].into(),
+            retry: true,
+            jid: crate::new_jid(),
+            created_at: 1337.0,
+            enqueued_at: None,
+            failed_at: None,
+            error_message: None,
+            retry_count: None,
+            retried_at: None,
+        }
+    }
+
+    #[derive(Clone)]
+    struct TestWorker {
+        touched: Arc<Mutex<bool>>,
+    }
+
+    #[async_trait]
+    impl Worker<()> for TestWorker {
+        async fn perform(&self, _args: ()) -> Result<(), Box<dyn std::error::Error>> {
+            *self.touched.lock().await = true;
+            Ok(())
+        }
+    }
+
+    #[tokio::test]
+    async fn calls_through_a_middleware_stack() {
+        let inner = Arc::new(TestWorker {
+            touched: Arc::new(Mutex::new(false)),
+        });
+        let worker = Arc::new(WorkerRef::wrap(Arc::clone(&inner)));
+
+        let job = job();
+        let mut chain = Chain::empty();
+        chain.using(Box::new(HandlerMiddleware)).await;
+        chain
+            .call(&job, worker.clone(), redis().await)
+            .await
+            .unwrap();
+
+        assert!(
+            *inner.touched.lock().await,
+            "The job was processed by the middleware",
+        );
+    }
+}
