@@ -1,13 +1,14 @@
-use crate::{Chain, Job, Scheduled, ServerMiddleware, UnitOfWork, Worker};
+use crate::{Chain, Job, Scheduled, ServerMiddleware, UnitOfWork, Worker, WorkerRef};
 use bb8_redis::{bb8::Pool, redis::AsyncCommands, RedisConnectionManager};
 use slog::{error, info};
 use std::collections::BTreeMap;
+use std::sync::Arc;
 
 #[derive(Clone)]
 pub struct Processor {
     redis: Pool<RedisConnectionManager>,
     queues: Vec<String>,
-    workers: BTreeMap<String, Box<dyn Worker>>,
+    workers: BTreeMap<String, Arc<WorkerRef>>,
     logger: slog::Logger,
     chain: Chain,
 }
@@ -51,7 +52,7 @@ impl Processor {
 
         if let Some(worker) = self.workers.get_mut(&work.job.class) {
             self.chain
-                .call(work.job.clone(), worker.clone(), self.redis.clone())
+                .call(&work.job, worker.clone(), self.redis.clone())
                 .await?;
         } else {
             error!(
@@ -78,8 +79,15 @@ impl Processor {
         Ok(())
     }
 
-    pub fn register<W: Worker + 'static>(&mut self, worker: W) {
-        self.workers.insert(W::class_name(), Box::new(worker));
+    pub fn register<
+        Args: Sync + Send + for<'de> serde::Deserialize<'de> + 'static,
+        W: Worker<Args> + 'static,
+    >(
+        &mut self,
+        worker: W,
+    ) {
+        self.workers
+            .insert(W::class_name(), Arc::new(WorkerRef::wrap(Arc::new(worker))));
     }
 
     /// Takes self to consume the processor. This is for life-cycle management, not
